/*
 * Copyright (C) 2008, 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSGlobalData.h"

#include "ArgList.h"
#include "Heap.h"
#include "CommonIdentifiers.h"
#include "DebuggerActivation.h"
#include "FunctionConstructor.h"
#include "GetterSetter.h"
#include "Interpreter.h"
#include "JSActivation.h"
#include "JSAPIValueWrapper.h"
#include "JSArray.h"
#include "JSByteArray.h"
#include "JSClassRef.h"
#include "JSFunction.h"
#include "JSLock.h"
#include "JSNotAnObject.h"
#include "JSPropertyNameIterator.h"
#include "JSStaticScopeObject.h"
#include "Lexer.h"
#include "Lookup.h"
#include "Nodes.h"
#include "Parser.h"
#include "RegExpCache.h"
#include "RegExpObject.h"
#include "StrictEvalActivation.h"
#include <wtf/WTFThreadData.h>
#if ENABLE(REGEXP_TRACING)
#include "RegExp.h"
#endif


#if ENABLE(JSC_MULTIPLE_THREADS)
#include <wtf/Threading.h>
#endif

#if PLATFORM(MAC)
#include "ProfilerServer.h"
#include <CoreFoundation/CoreFoundation.h>
#endif

using namespace WTF;

namespace {

using namespace JSC;

class Recompiler : public MarkedBlock::VoidFunctor {
public:
    void operator()(JSCell*);
};

inline void Recompiler::operator()(JSCell* cell)
{
    if (!cell->inherits(&JSFunction::s_info))
        return;
    JSFunction* function = asFunction(cell);
    if (!function->executable() || function->executable()->isHostFunction())
        return;
    function->jsExecutable()->discardCode();
}

} // namespace

namespace JSC {

extern JSC_CONST_HASHTABLE HashTable arrayConstructorTable;
extern JSC_CONST_HASHTABLE HashTable arrayPrototypeTable;
extern JSC_CONST_HASHTABLE HashTable booleanPrototypeTable;
extern JSC_CONST_HASHTABLE HashTable jsonTable;
extern JSC_CONST_HASHTABLE HashTable dateTable;
extern JSC_CONST_HASHTABLE HashTable dateConstructorTable;
extern JSC_CONST_HASHTABLE HashTable errorPrototypeTable;
extern JSC_CONST_HASHTABLE HashTable globalObjectTable;
extern JSC_CONST_HASHTABLE HashTable mathTable;
extern JSC_CONST_HASHTABLE HashTable numberConstructorTable;
extern JSC_CONST_HASHTABLE HashTable numberPrototypeTable;
extern JSC_CONST_HASHTABLE HashTable objectConstructorTable;
extern JSC_CONST_HASHTABLE HashTable objectPrototypeTable;
extern JSC_CONST_HASHTABLE HashTable regExpTable;
extern JSC_CONST_HASHTABLE HashTable regExpConstructorTable;
extern JSC_CONST_HASHTABLE HashTable regExpPrototypeTable;
extern JSC_CONST_HASHTABLE HashTable stringTable;
extern JSC_CONST_HASHTABLE HashTable stringConstructorTable;

void* JSGlobalData::jsFinalObjectVPtr;
void* JSGlobalData::jsArrayVPtr;
void* JSGlobalData::jsByteArrayVPtr;
void* JSGlobalData::jsStringVPtr;
void* JSGlobalData::jsFunctionVPtr;

#if COMPILER(GCC)
// Work around for gcc trying to coalesce our reads of the various cell vptrs
#define CLOBBER_MEMORY() do { \
    asm volatile ("" : : : "memory"); \
} while (false)
#else
#define CLOBBER_MEMORY() do { } while (false)
#endif

void JSGlobalData::storeVPtrs()
{
    // Enough storage to fit a JSArray, JSByteArray, JSString, or JSFunction.
    // COMPILE_ASSERTS below check that this is true.
    char storage[64];

    COMPILE_ASSERT(sizeof(JSFinalObject) <= sizeof(storage), sizeof_JSFinalObject_must_be_less_than_storage);
    JSCell* jsFinalObject = new (storage) JSFinalObject(JSFinalObject::VPtrStealingHack);
    CLOBBER_MEMORY();
    JSGlobalData::jsFinalObjectVPtr = jsFinalObject->vptr();

    COMPILE_ASSERT(sizeof(JSArray) <= sizeof(storage), sizeof_JSArray_must_be_less_than_storage);
    JSCell* jsArray = new (storage) JSArray(JSArray::VPtrStealingHack);
    CLOBBER_MEMORY();
    JSGlobalData::jsArrayVPtr = jsArray->vptr();

    COMPILE_ASSERT(sizeof(JSByteArray) <= sizeof(storage), sizeof_JSByteArray_must_be_less_than_storage);
    JSCell* jsByteArray = new (storage) JSByteArray(JSByteArray::VPtrStealingHack);
    CLOBBER_MEMORY();
    JSGlobalData::jsByteArrayVPtr = jsByteArray->vptr();

    COMPILE_ASSERT(sizeof(JSString) <= sizeof(storage), sizeof_JSString_must_be_less_than_storage);
    JSCell* jsString = new (storage) JSString(JSString::VPtrStealingHack);
    CLOBBER_MEMORY();
    JSGlobalData::jsStringVPtr = jsString->vptr();

    COMPILE_ASSERT(sizeof(JSFunction) <= sizeof(storage), sizeof_JSFunction_must_be_less_than_storage);
    JSCell* jsFunction = new (storage) JSFunction(JSCell::VPtrStealingHack);
    CLOBBER_MEMORY();
    JSGlobalData::jsFunctionVPtr = jsFunction->vptr();
}

JSGlobalData::JSGlobalData(GlobalDataType globalDataType, ThreadStackType threadStackType, HeapSize heapSize)
    : globalDataType(globalDataType)
    , clientData(0)
    , topCallFrame(CallFrame::noCaller())
    , arrayConstructorTable(fastNew<HashTable>(JSC::arrayConstructorTable))
    , arrayPrototypeTable(fastNew<HashTable>(JSC::arrayPrototypeTable))
    , booleanPrototypeTable(fastNew<HashTable>(JSC::booleanPrototypeTable))
    , dateTable(fastNew<HashTable>(JSC::dateTable))
    , dateConstructorTable(fastNew<HashTable>(JSC::dateConstructorTable))
    , errorPrototypeTable(fastNew<HashTable>(JSC::errorPrototypeTable))
    , globalObjectTable(fastNew<HashTable>(JSC::globalObjectTable))
    , jsonTable(fastNew<HashTable>(JSC::jsonTable))
    , mathTable(fastNew<HashTable>(JSC::mathTable))
    , numberConstructorTable(fastNew<HashTable>(JSC::numberConstructorTable))
    , numberPrototypeTable(fastNew<HashTable>(JSC::numberPrototypeTable))
    , objectConstructorTable(fastNew<HashTable>(JSC::objectConstructorTable))
    , objectPrototypeTable(fastNew<HashTable>(JSC::objectPrototypeTable))
    , regExpTable(fastNew<HashTable>(JSC::regExpTable))
    , regExpConstructorTable(fastNew<HashTable>(JSC::regExpConstructorTable))
    , regExpPrototypeTable(fastNew<HashTable>(JSC::regExpPrototypeTable))
    , stringTable(fastNew<HashTable>(JSC::stringTable))
    , stringConstructorTable(fastNew<HashTable>(JSC::stringConstructorTable))
    , identifierTable(globalDataType == Default ? wtfThreadData().currentIdentifierTable() : createIdentifierTable())
    , propertyNames(new CommonIdentifiers(this))
    , emptyList(new MarkedArgumentBuffer)
#if ENABLE(ASSEMBLER)
    , executableAllocator(*this)
#endif
    , lexer(new Lexer(this))
    , parser(new Parser)
    , interpreter(0)
    , heap(this, heapSize)
    , dynamicGlobalObject(0)
    , cachedUTCOffset(std::numeric_limits<double>::quiet_NaN())
    , maxReentryDepth(threadStackType == ThreadStackTypeSmall ? MaxSmallThreadReentryDepth : MaxLargeThreadReentryDepth)
    , m_regExpCache(new RegExpCache(this))
#if ENABLE(REGEXP_TRACING)
    , m_rtTraceList(new RTTraceList())
#endif
#ifndef NDEBUG
    , exclusiveThread(0)
#endif
#if ENABLE(GC_VALIDATION)
    , m_isInitializingObject(false)
#endif
{
    interpreter = new Interpreter;
    if (globalDataType == Default)
        m_stack = wtfThreadData().stack();

    // Need to be careful to keep everything consistent here
    IdentifierTable* existingEntryIdentifierTable = wtfThreadData().setCurrentIdentifierTable(identifierTable);
    JSLock lock(SilenceAssertionsOnly);
    structureStructure.set(*this, Structure::createStructure(*this));
    debuggerActivationStructure.set(*this, DebuggerActivation::createStructure(*this, 0, jsNull()));
    activationStructure.set(*this, JSActivation::createStructure(*this, 0, jsNull()));
    interruptedExecutionErrorStructure.set(*this, InterruptedExecutionError::createStructure(*this, 0, jsNull()));
    terminatedExecutionErrorStructure.set(*this, TerminatedExecutionError::createStructure(*this, 0, jsNull()));
    staticScopeStructure.set(*this, JSStaticScopeObject::createStructure(*this, 0, jsNull()));
    strictEvalActivationStructure.set(*this, StrictEvalActivation::createStructure(*this, 0, jsNull()));
    stringStructure.set(*this, JSString::createStructure(*this, 0, jsNull()));
    notAnObjectStructure.set(*this, JSNotAnObject::createStructure(*this, 0, jsNull()));
    propertyNameIteratorStructure.set(*this, JSPropertyNameIterator::createStructure(*this, 0, jsNull()));
    getterSetterStructure.set(*this, GetterSetter::createStructure(*this, 0, jsNull()));
    apiWrapperStructure.set(*this, JSAPIValueWrapper::createStructure(*this, 0, jsNull()));
    scopeChainNodeStructure.set(*this, ScopeChainNode::createStructure(*this, 0, jsNull()));
    executableStructure.set(*this, ExecutableBase::createStructure(*this, 0, jsNull()));
    nativeExecutableStructure.set(*this, NativeExecutable::createStructure(*this, 0, jsNull()));
    evalExecutableStructure.set(*this, EvalExecutable::createStructure(*this, 0, jsNull()));
    programExecutableStructure.set(*this, ProgramExecutable::createStructure(*this, 0, jsNull()));
    functionExecutableStructure.set(*this, FunctionExecutable::createStructure(*this, 0, jsNull()));
    regExpStructure.set(*this, RegExp::createStructure(*this, 0, jsNull()));
    structureChainStructure.set(*this, StructureChain::createStructure(*this, 0, jsNull()));

    wtfThreadData().setCurrentIdentifierTable(existingEntryIdentifierTable);

#if PLATFORM(MAC)
    startProfilerServerIfNeeded();
#endif
#if ENABLE(JIT) && ENABLE(INTERPRETER)
#if USE(CF)
    CFStringRef canUseJITKey = CFStringCreateWithCString(0 , "JavaScriptCoreUseJIT", kCFStringEncodingMacRoman);
    CFBooleanRef canUseJIT = (CFBooleanRef)CFPreferencesCopyAppValue(canUseJITKey, kCFPreferencesCurrentApplication);
    if (canUseJIT) {
        m_canUseJIT = kCFBooleanTrue == canUseJIT;
        CFRelease(canUseJIT);
    } else {
      char* canUseJITString = getenv("JavaScriptCoreUseJIT");
      m_canUseJIT = !canUseJITString || atoi(canUseJITString);
    }
    CFRelease(canUseJITKey);
#elif OS(UNIX)
    char* canUseJITString = getenv("JavaScriptCoreUseJIT");
    m_canUseJIT = !canUseJITString || atoi(canUseJITString);
#else
    m_canUseJIT = true;
#endif
#endif
#if ENABLE(JIT)

#if PLATFORM(WIIU)
    m_execCnt = 0;
	m_myThread = pthread_self();

	pthread_mutex_lock(&s_execGDMutex);
	s_execGDList.push_back(this);
	pthread_mutex_unlock(&s_execGDMutex);
#endif

#if ENABLE(INTERPRETER)
    if (m_canUseJIT)
        m_canUseJIT = executableAllocator.isValid();
#endif
    jitStubs = adoptPtr(new JITThunks(this));
#endif

    heap.notifyIsSafeToCollect();
}

void JSGlobalData::clearBuiltinStructures()
{
    structureStructure.clear();
    debuggerActivationStructure.clear();
    activationStructure.clear();
    interruptedExecutionErrorStructure.clear();
    terminatedExecutionErrorStructure.clear();
    staticScopeStructure.clear();
    strictEvalActivationStructure.clear();
    stringStructure.clear();
    notAnObjectStructure.clear();
    propertyNameIteratorStructure.clear();
    getterSetterStructure.clear();
    apiWrapperStructure.clear();
    scopeChainNodeStructure.clear();
    executableStructure.clear();
    nativeExecutableStructure.clear();
    evalExecutableStructure.clear();
    programExecutableStructure.clear();
    functionExecutableStructure.clear();
    regExpStructure.clear();
    structureChainStructure.clear();
}

JSGlobalData::~JSGlobalData()
{
    // By the time this is destroyed, heap.destroy() must already have been called.

    delete interpreter;
#ifndef NDEBUG
    // Zeroing out to make the behavior more predictable when someone attempts to use a deleted instance.
    interpreter = 0;
#endif

    arrayPrototypeTable->deleteTable();
    arrayConstructorTable->deleteTable();
    booleanPrototypeTable->deleteTable();
    dateTable->deleteTable();
    dateConstructorTable->deleteTable();
    errorPrototypeTable->deleteTable();
    globalObjectTable->deleteTable();
    jsonTable->deleteTable();
    mathTable->deleteTable();
    numberConstructorTable->deleteTable();
    numberPrototypeTable->deleteTable();
    objectConstructorTable->deleteTable();
    objectPrototypeTable->deleteTable();
    regExpTable->deleteTable();
    regExpConstructorTable->deleteTable();
    regExpPrototypeTable->deleteTable();
    stringTable->deleteTable();
    stringConstructorTable->deleteTable();

    fastDelete(const_cast<HashTable*>(arrayConstructorTable));
    fastDelete(const_cast<HashTable*>(arrayPrototypeTable));
    fastDelete(const_cast<HashTable*>(booleanPrototypeTable));
    fastDelete(const_cast<HashTable*>(dateTable));
    fastDelete(const_cast<HashTable*>(dateConstructorTable));
    fastDelete(const_cast<HashTable*>(errorPrototypeTable));
    fastDelete(const_cast<HashTable*>(globalObjectTable));
    fastDelete(const_cast<HashTable*>(jsonTable));
    fastDelete(const_cast<HashTable*>(mathTable));
    fastDelete(const_cast<HashTable*>(numberConstructorTable));
    fastDelete(const_cast<HashTable*>(numberPrototypeTable));
    fastDelete(const_cast<HashTable*>(objectConstructorTable));
    fastDelete(const_cast<HashTable*>(objectPrototypeTable));
    fastDelete(const_cast<HashTable*>(regExpTable));
    fastDelete(const_cast<HashTable*>(regExpConstructorTable));
    fastDelete(const_cast<HashTable*>(regExpPrototypeTable));
    fastDelete(const_cast<HashTable*>(stringTable));
    fastDelete(const_cast<HashTable*>(stringConstructorTable));

    delete parser;
    delete lexer;

    deleteAllValues(opaqueJSClassData);

    delete emptyList;

    delete propertyNames;
    if (globalDataType != Default)
        deleteIdentifierTable(identifierTable);

    delete clientData;
    delete m_regExpCache;
#if ENABLE(REGEXP_TRACING)
    delete m_rtTraceList;
#endif

#if PLATFORM(WIIU) && ENABLE(JIT)
	pthread_mutex_lock(&s_execGDMutex);
	s_execGDList.remove(this);
	pthread_mutex_unlock(&s_execGDMutex);
#endif
}

#if PLATFORM(WIIU) && ENABLE(JIT)
/*
    The "JIT Locker"

    At times we need to lock down any use of JIT on WiiU. Unfortunately we cannot fallback
    to interpreted JS (at least on this version of WebKit). Hence we need to avoid calling
    any JIT code during these intervals.
    The code below will block any thread attempting to execute JIT, but effectively stopping
    it untile JIT execution is possible again.
    Additonal complexity: we need to pool for state changes if we should block on JIT execution
    on the main WebKit thread. Hence we supply a callback, which can execute the needed things
    without breaking the abstraction layers too much.

	In addition this takes care of coordinating code generation vs. execution with >1 threads.

	Note:

	A Webworker, which is continously inside JIT code for long periods of time will unfortunately
	stall process switches while it does so... but that cannot be solved unless we could ru JIT
	on more then one core... (making much of this locking unneccessary)
*/

bool						JSGlobalData::s_bLockJIT;                               //!< True if JIT is currently locked, false otherwise

pthread_t				    JSGlobalData::s_execMainThread;                         //!< Main thread handle to allow identification of it (see above why we need this)
void                        (*JSGlobalData::s_execProcessSwitchPollCallback)(void); //!< Callback to allow for polling etc. while waiting for JIT access (see above why we need this)

pthread_mutex_t			    JSGlobalData::s_execAccessMutex;                        //!< Mutex to protect access to exec / codegen state
volatile int				JSGlobalData::s_codegenCnt;								//!< Number of currently active codegen operations
volatile int        		JSGlobalData::s_execCnt;								//!< Number of currently active exec operations
pthread_cond_t				JSGlobalData::s_codegenCond;							//!< Conditional variable to signal codegen operations being possibly down to zero
pthread_cond_t				JSGlobalData::s_execCond;								//!< Conditional variable to signal exec operations being possibly down to zero
std::list<JSGlobalData*>	JSGlobalData::s_execGDList;								//!< List of JSGlobalData instances to be able to find each threads instance from a static call
pthread_mutex_t				JSGlobalData::s_execGDMutex;							//!< Access control for JSGlobalData list

//! Initialize JIT locker code
void JSGlobalData::initJITLocker(void (*execProcessSwitchPollCallback)(void))
{
    pthread_mutex_init(&s_execAccessMutex, NULL);
	pthread_cond_init(&s_execCond, NULL);
	pthread_cond_init(&s_codegenCond, NULL);
	s_codegenCnt = 0;
	s_execCnt = 0;

    pthread_mutex_init(&s_execGDMutex, NULL);

    s_execMainThread = pthread_self();
    s_execProcessSwitchPollCallback = execProcessSwitchPollCallback;
}

//! Shutdown JIT locker code
void JSGlobalData::exitJITLocker()
{
    pthread_mutex_destroy(&s_execAccessMutex);
	pthread_cond_destroy(&s_execCond);
	pthread_cond_destroy(&s_codegenCond);
    pthread_mutex_destroy(&s_execGDMutex);
}

//! Set new state to JIT locker
void JSGlobalData::setLockJIT(bool bLockJIT)
{
	// note: we assume only ONE thread will ever call this at any given time!
    if (s_bLockJIT != bLockJIT)
    {
        s_bLockJIT = bLockJIT;

		if (bLockJIT)
		{
			// Lock down JIT...

			pthread_mutex_lock(&s_execAccessMutex);

			// We pretend to do both codegen & execute work to fake any new JIT related work to be blocked...
			++s_codegenCnt;
			++s_execCnt;

			// Wait for all codegens to be done (but our fake one)
			while(s_codegenCnt > 1)
				pthread_cond_wait(&s_codegenCond, &s_execAccessMutex);

			// Wait for all current executes to be done (but our fake one)
			while(s_execCnt > 1)
				pthread_cond_wait(&s_execCond, &s_execAccessMutex);

			// Switch off codegen in case we still had it on before
			// (the codegenEnd() calls outside this would not have done it as they see s_codegenCnt as not being zero)
			OS::SwitchSecCodeGenMode(true);

			ASSERT(s_codegenCnt == 1);
			ASSERT(s_execCnt == 1);

			pthread_mutex_unlock(&s_execAccessMutex);
		}
		else
		{
			// Unlock JIT...

			pthread_mutex_lock(&s_execAccessMutex);
			// We're no longer pretending to keep things busy...
			ASSERT(s_codegenCnt == 1);
			ASSERT(s_execCnt == 1);
			s_codegenCnt = 0;
			s_execCnt = 0;
			// Kick any waiting threads...
			pthread_cond_signal(&s_codegenCond);
			pthread_cond_signal(&s_execCond);
			pthread_mutex_unlock(&s_execAccessMutex);
		}
    }
    pthread_mutex_unlock(&s_execAccessMutex);
}

//! Function to be called prior to any JIT execution
void JSGlobalData::executeBegin()
{
    if (m_execCnt++ == 0)
    {
        if (s_execProcessSwitchPollCallback && pthread_equal(pthread_self(), s_execMainThread))
        {
			pthread_mutex_lock(&s_execAccessMutex);
			while(s_codegenCnt)
            {
				// Wait for the condvar to be triggered for 15ms from now on...
				struct timeval	currentTime;
				struct timespec	targetTime;
				gettimeofday(&currentTime, NULL);
				targetTime.tv_sec = currentTime.tv_sec;
				targetTime.tv_nsec = currentTime.tv_usec * 1000 + 15 * 1000 * 1000;
				if (targetTime.tv_nsec >= 1000 * 1000 * 1000)
				{
					targetTime.tv_nsec -= 1000 * 1000 * 1000;
					targetTime.tv_sec += 1;
				}
				pthread_cond_timedwait(&s_codegenCond, &s_execAccessMutex, &targetTime);

				if (s_codegenCnt == 0)
					break;

				// We need to release the lock while we handle the callback. It may itself call a function in need of the mutex...
				pthread_mutex_unlock(&s_execAccessMutex);

				// Execute process switch callback
                s_execProcessSwitchPollCallback();

				pthread_mutex_lock(&s_execAccessMutex);
            }
			++s_execCnt;
			pthread_mutex_unlock(&s_execAccessMutex);
		}
		else
		{
			// Not on the main webkit thread / no callback known: we can do this the easy way...
			pthread_mutex_lock(&s_execAccessMutex);
			while(s_codegenCnt)
				pthread_cond_wait(&s_codegenCond, &s_execAccessMutex);
			++s_execCnt;
			pthread_mutex_unlock(&s_execAccessMutex);
		}
    }
}

//! Function to be called after JIT execution
void JSGlobalData::executeEnd()
{
	// End execution for this thread
    if (--m_execCnt == 0)
	{
		pthread_mutex_lock(&s_execAccessMutex);
		--s_execCnt;
		pthread_cond_signal(&s_execCond);
		pthread_mutex_unlock(&s_execAccessMutex);
	}
}

//! Function used to sync up with other threads that might execute code before codegen begins
void JSGlobalData::codegenBegin()
{
	// Force down the global execute counter if this context should be executing right now...
	pthread_t thisThread = pthread_self();
	pthread_mutex_lock(&s_execGDMutex);
	for(std::list<JSGlobalData*>::const_iterator it=s_execGDList.begin(); it!=s_execGDList.end(); ++it)
	{
		// This us?
		JSGlobalData *pJSGD = *it;
		if (pthread_equal(pJSGD->m_myThread, thisThread))
		{
			// Ok. Are we in execution mode right now?
			if (pJSGD->m_execCnt)
			{
				// Yes. Remove our count from global counter for now...
				pthread_mutex_lock(&s_execAccessMutex);
				--s_execCnt;
				pthread_cond_signal(&s_execCond);
				pthread_mutex_unlock(&s_execAccessMutex);
			}
		break;
		}
	}
	pthread_mutex_unlock(&s_execGDMutex);

	// Now wait for (all other) execution to stop before we allow codegen...
	if (s_execProcessSwitchPollCallback && pthread_equal(pthread_self(), s_execMainThread))
	{
		pthread_mutex_lock(&s_execAccessMutex);
		while(s_execCnt)
		{
			// Wait for the condvar to be triggered for 15ms from now on...
			struct timeval	currentTime;
			struct timespec	targetTime;
			gettimeofday(&currentTime, NULL);
			targetTime.tv_sec = currentTime.tv_sec;
			targetTime.tv_nsec = currentTime.tv_usec * 1000 + 15 * 1000 * 1000;
			if (targetTime.tv_nsec >= 1000 * 1000 * 1000)
			{
				targetTime.tv_nsec -= 1000 * 1000 * 1000;
				targetTime.tv_sec += 1;
			}
			pthread_cond_timedwait(&s_execCond, &s_execAccessMutex, &targetTime);

			if (s_execCnt == 0)
				break;

			// We need to release the lock while we handle the callback. It may itself call a function in need of the mutex...
			pthread_mutex_unlock(&s_execAccessMutex);

			// Execute process switch callback
			s_execProcessSwitchPollCallback();

			pthread_mutex_lock(&s_execAccessMutex);
		}
		if (s_codegenCnt++ == 0)
			OS::SwitchSecCodeGenMode(false);
		pthread_mutex_unlock(&s_execAccessMutex);
	}
	else
	{
		// Not on the main webkit thread / no callback known: we can do this the easy way...
		pthread_mutex_lock(&s_execAccessMutex);
		while(s_execCnt)
			pthread_cond_wait(&s_execCond, &s_execAccessMutex);
		if (s_codegenCnt++ == 0)
			OS::SwitchSecCodeGenMode(false);
		pthread_mutex_unlock(&s_execAccessMutex);
	}
}

//! Function used to allow code execution again after codegen ends
void JSGlobalData::codegenEnd()
{
	pthread_mutex_lock(&s_execAccessMutex);
	if (--s_codegenCnt == 0)
		OS::SwitchSecCodeGenMode(true);
	pthread_cond_signal(&s_codegenCond);
	pthread_mutex_unlock(&s_execAccessMutex);

	// Begin execution again if we had it going before...
	// (this may stall until any other codegen threads have stopped - they may have started while we had been doing codegen!)
	pthread_t thisThread = pthread_self();
	JSGlobalData *pJSGD = NULL;
	pthread_mutex_lock(&s_execGDMutex);
	for(std::list<JSGlobalData*>::const_iterator it=s_execGDList.begin(); it!=s_execGDList.end(); ++it)
	{
		if (pthread_equal((*it)->m_myThread, thisThread))
		{
			pJSGD = *it;
			break;
		}
	}
	pthread_mutex_unlock(&s_execGDMutex);
	// Did our context execute anything?
	int execCnt = pJSGD->m_execCnt;
	if (execCnt)
	{
		// Yes. Cheat executeBegin into doing a proper "begin" every time...
		pJSGD->m_execCnt = 0;
		pJSGD->executeBegin();
		pJSGD->m_execCnt = execCnt;
	}
}

void JSGlobalData::SetCodeGenMode(bool bWritable)
{
    if (bWritable)
		codegenBegin();
    else
		codegenEnd();
}

#endif

PassRefPtr<JSGlobalData> JSGlobalData::createContextGroup(ThreadStackType type, HeapSize heapSize)
{
    return adoptRef(new JSGlobalData(APIContextGroup, type, heapSize));
}

PassRefPtr<JSGlobalData> JSGlobalData::create(ThreadStackType type, HeapSize heapSize)
{
    return adoptRef(new JSGlobalData(Default, type, heapSize));
}

PassRefPtr<JSGlobalData> JSGlobalData::createLeaked(ThreadStackType type, HeapSize heapSize)
{
    return create(type, heapSize);
}

bool JSGlobalData::sharedInstanceExists()
{
    return sharedInstanceInternal();
}

JSGlobalData& JSGlobalData::sharedInstance()
{
    JSGlobalData*& instance = sharedInstanceInternal();
    if (!instance) {
        instance = adoptRef(new JSGlobalData(APIShared, ThreadStackTypeSmall, SmallHeap)).leakRef();
#if ENABLE(JSC_MULTIPLE_THREADS)
        instance->makeUsableFromMultipleThreads();
#endif
    }
    return *instance;
}

JSGlobalData*& JSGlobalData::sharedInstanceInternal()
{
    ASSERT(JSLock::currentThreadIsHoldingLock());
    static JSGlobalData* sharedInstance;
    return sharedInstance;
}

#if ENABLE(JIT)
NativeExecutable* JSGlobalData::getHostFunction(NativeFunction function)
{
    return jitStubs->hostFunctionStub(this, function);
}
NativeExecutable* JSGlobalData::getHostFunction(NativeFunction function, ThunkGenerator generator)
{
    return jitStubs->hostFunctionStub(this, function, generator);
}
#else
NativeExecutable* JSGlobalData::getHostFunction(NativeFunction function)
{
    return NativeExecutable::create(*this, function, callHostFunctionAsConstructor);
}
#endif

JSGlobalData::ClientData::~ClientData()
{
}

void JSGlobalData::resetDateCache()
{
    cachedUTCOffset = std::numeric_limits<double>::quiet_NaN();
    dstOffsetCache.reset();
    cachedDateString = UString();
    cachedDateStringValue = std::numeric_limits<double>::quiet_NaN();
    dateInstanceCache.reset();
}

void JSGlobalData::startSampling()
{
    interpreter->startSampling();
}

void JSGlobalData::stopSampling()
{
    interpreter->stopSampling();
}

void JSGlobalData::dumpSampleData(ExecState* exec)
{
    interpreter->dumpSampleData(exec);
#if ENABLE(ASSEMBLER)
    ExecutableAllocator::dumpProfile();
#endif
}

void JSGlobalData::recompileAllJSFunctions()
{
    // If JavaScript is running, it's not safe to recompile, since we'll end
    // up throwing away code that is live on the stack.
    ASSERT(!dynamicGlobalObject);
    
    heap.forEachCell<Recompiler>();
}

struct StackPreservingRecompiler : public MarkedBlock::VoidFunctor {
    HashSet<FunctionExecutable*> currentlyExecutingFunctions;
    void operator()(JSCell* cell)
    {
        if (!cell->inherits(&FunctionExecutable::s_info))
            return;
        FunctionExecutable* executable = static_cast<FunctionExecutable*>(cell);
        if (currentlyExecutingFunctions.contains(executable))
            return;
        executable->discardCode();
    }
};

void JSGlobalData::releaseExecutableMemory()
{
    if (dynamicGlobalObject) {
        StackPreservingRecompiler recompiler;
        HashSet<JSCell*> roots;
        heap.getConservativeRegisterRoots(roots);
        HashSet<JSCell*>::iterator end = roots.end();
        for (HashSet<JSCell*>::iterator ptr = roots.begin(); ptr != end; ++ptr) {
            ScriptExecutable* executable = 0;
            JSCell* cell = *ptr;
            if (cell->inherits(&ScriptExecutable::s_info))
                executable = static_cast<ScriptExecutable*>(*ptr);
            else if (cell->inherits(&JSFunction::s_info)) {
                JSFunction* function = asFunction(*ptr);
                if (function->isHostFunction())
                    continue;
                executable = function->jsExecutable();
            } else
                continue;
            ASSERT(executable->inherits(&ScriptExecutable::s_info));
            executable->unlinkCalls();
            if (executable->inherits(&FunctionExecutable::s_info))
                recompiler.currentlyExecutingFunctions.add(static_cast<FunctionExecutable*>(executable));
                
        }
        heap.forEachCell<StackPreservingRecompiler>(recompiler);
    }
    m_regExpCache->invalidateCode();
    heap.collectAllGarbage();
}
    
void releaseExecutableMemory(JSGlobalData& globalData)
{
    globalData.releaseExecutableMemory();
}

#if ENABLE(REGEXP_TRACING)
void JSGlobalData::addRegExpToTrace(RegExp* regExp)
{
    m_rtTraceList->add(regExp);
}

void JSGlobalData::dumpRegExpTrace()
{
    // The first RegExp object is ignored.  It is create by the RegExpPrototype ctor and not used.
    RTTraceList::iterator iter = ++m_rtTraceList->begin();
    
    if (iter != m_rtTraceList->end()) {
        printf("\nRegExp Tracing\n");
        printf("                                                            match()    matches\n");
        printf("Regular Expression                          JIT Address      calls      found\n");
        printf("----------------------------------------+----------------+----------+----------\n");
    
        unsigned reCount = 0;
    
        for (; iter != m_rtTraceList->end(); ++iter, ++reCount)
            (*iter)->printTraceData();

        printf("%d Regular Expressions\n", reCount);
    }
    
    m_rtTraceList->clear();
}
#else
void JSGlobalData::dumpRegExpTrace()
{
}
#endif

} // namespace JSC
