/*
 * Copyright (C) 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "FontCache.h"
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include "NotImplemented.h"
#include <PlatformOS.h>
using namespace LibPlatform;
#endif

#include "Font.h"
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
#include "OwnPtrCairo.h"
#include "RefPtrCairo.h"
#endif
#include "SimpleFontData.h"

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
#include <cairo-ft.h>
#include <cairo.h>
#include <fontconfig/fcfreetype.h>
#endif
#include <wtf/Assertions.h>
#include <wtf/text/CString.h>

#include "efontconfig/efontconfig.h"
#if !PLATFORM(WIIU)
#include "renderer/WiiURenderer.h"
#else
#define LOG_SIMPLE_RENDER_MSG(vMessage) {}
#define LOG_RENDER_MSG(vMessage, ...) {}
#endif

namespace WebCore {

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
FcPattern* createFontConfigPatternForCharacters(const UChar* characters, int length)
{
    FcPattern* pattern = FcPatternCreate();

    FcCharSet* fontConfigCharSet = FcCharSetCreate();
    for (int i = 0; i < length; ++i) {
        if (U16_IS_SURROGATE(characters[i]) && U16_IS_SURROGATE_LEAD(characters[i])
                && i != length - 1 && U16_IS_TRAIL(characters[i + 1])) {
            FcCharSetAddChar(fontConfigCharSet, U16_GET_SUPPLEMENTARY(characters[i], characters[i+1]));
            i++;
        } else
            FcCharSetAddChar(fontConfigCharSet, characters[i]);
    }
    FcPatternAddCharSet(pattern, FC_CHARSET, fontConfigCharSet);
    FcCharSetDestroy(fontConfigCharSet);

    FcPatternAddBool(pattern, FC_SCALABLE, FcTrue);
    FcConfigSubstitute(0, pattern, FcMatchPattern);
    FcDefaultSubstitute(pattern);
    return pattern;
}
#else
void* createFontConfigPatternForCharacters(const UChar* characters, int length) {
	notImplemented();
    OS::Report("createFontConfigPatternForCharacters");
	return NULL;
}
#endif

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
FcPattern* findBestFontGivenFallbacks(const FontPlatformData& fontData, FcPattern* pattern)
{
    if (!fontData.m_pattern)
        return 0;

    if (!fontData.m_fallbacks) {
        FcResult fontConfigResult;
        fontData.m_fallbacks = FcFontSort(0, fontData.m_pattern.get(), FcTrue, 0, &fontConfigResult);
    }

    if (!fontData.m_fallbacks)
        return 0;

    FcFontSet* sets[] = { fontData.m_fallbacks };
    FcResult fontConfigResult;
    return FcFontSetMatch(0, sets, 1, pattern, &fontConfigResult);
}
#else
void* findBestFontGivenFallbacks(const FontPlatformData& fontData, void* pattern)
{
	notImplemented();
    OS::Report("findBestFontGivenFallbacks");
	return 0;
}
#endif

SimpleFontData* FontCache::getSimilarFontPlatformData(const Font& font)
{
    SimpleFontData* simpleFontData = 0;

    if (!simpleFontData)
       simpleFontData = getCachedFontData(font.fontDescription(), AtomicString("serif"));

    return simpleFontData;
}

SimpleFontData* FontCache::getLastResortFallbackFont(const FontDescription& fontDescription, ShouldRetain shouldRetain)
{
    // We want to return a fallback font here, otherwise the logic preventing FontConfig
    // matches for non-fallback fonts might return 0. See isFallbackFontAllowed.
    static AtomicString timesStr("serif");
    return getCachedFontData(fontDescription, timesStr, false, shouldRetain);
}

void FontCache::getTraitsInFamily(const AtomicString& familyName, Vector<unsigned>& traitsMasks)
{
}

int fontWeightToFontconfigWeight(FontWeight weight)
{
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    switch (weight) {
    case FontWeight100:
        return FC_WEIGHT_THIN;
    case FontWeight200:
        return FC_WEIGHT_ULTRALIGHT;
    case FontWeight300:
        return FC_WEIGHT_LIGHT;
    case FontWeight400:
        return FC_WEIGHT_REGULAR;
    case FontWeight500:
        return FC_WEIGHT_MEDIUM;
    case FontWeight600:
        return FC_WEIGHT_SEMIBOLD;
    case FontWeight700:
        return FC_WEIGHT_BOLD;
    case FontWeight800:
        return FC_WEIGHT_EXTRABOLD;
    case FontWeight900:
        return FC_WEIGHT_ULTRABLACK;
    default:
        ASSERT_NOT_REACHED();
        return FC_WEIGHT_REGULAR;
    }
#else
    notImplemented();
    return 0;
#endif
}

}
