/*
 * Copyright (c) 2011,2012 ACCESS CO., LTD. All rights reserved.
 * Copyright (c) 2013 NINTENDO.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 * 
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 */

#include "config.h"
#include "HTTPCache.h"
#include "SharedBuffer.h"
#include "KURL.h"
#include "CString.h"
#include "FileSystem.h"
#include "ResourceResponse.h"
#include "ResourceHandleClient.h"
#include <wtf/MathExtras.h>
#include <wtf/MD5.h>
#include <StringBuilder.h>

#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <shim.h>

#define FAT_HEADER_SIZE (sizeof(long long)*2 + sizeof(int) + sizeof(int) + sizeof(int) + sizeof(double)*4)


namespace WebCore {

// HTTPCacheResource

HTTPCachedResource::HTTPCachedResource()
{
    m_used = false;
    m_resourceData = 0;
}

HTTPCachedResource::HTTPCachedResource(KURL &url, ResourceResponse &response)
{
    m_used = false;

    m_url = url.string();
    setResourceResponse(response);
}


HTTPCachedResource::~HTTPCachedResource()
{
    if (m_resourceData) {
        m_resourceData = 0;
    }
}

// copy from CachedResource::freshnessLifetime()
double HTTPCachedResource::freshnessLifetime()
{
    // RFC2616 13.2.4
    if (isfinite(m_maxAge)) {
        return m_maxAge;
    }

    if (isfinite(m_expires)) {
        return m_expires - m_date;
    }

    if (isfinite(m_lastModified)) {
        return (m_date - m_lastModified) * 0.1;
    }

    // If no cache headers are present, the specification leaves the decision to the UA. Other browsers seem to opt for 0.
    return 0;
}

bool HTTPCachedResource::isExpired()
{
    return (currentTime() - m_date > freshnessLifetime());
}

bool HTTPCachedResource::needRevalidate(double override)
{
    if (override != -1) {
        //check for override...
        //OS::Report("override: %f, current: %f, date: %f, delta: %f %d\n", override, currentTime(), m_date, currentTime() - m_date, (currentTime() - m_date) >= 0.0);
        return m_mustRevalidate || m_noCache || (currentTime() - m_date) >= override; 
    } else { 
        return m_mustRevalidate || m_noCache || isExpired();
    }
}

void HTTPCachedResource::update(bool noCache, bool mustRevalidate, double expires, double maxAge)
{
    m_noCache = noCache;
    m_mustRevalidate = mustRevalidate;

    if (isfinite(maxAge)) {
        if (!isfinite(m_maxAge) || (m_maxAge > maxAge))
            m_maxAge = maxAge;
    }
    if (isfinite(expires)) {
        if (!isfinite(m_expires) || (m_expires > expires))
            m_expires = expires;
    }
}

bool HTTPCachedResource::setResourceResponse(ResourceResponse &response)
{
    m_mimeType = response.mimeType();
    m_expectedContentLength = response.expectedContentLength();
    m_textEncodingName = response.textEncodingName();
    m_suggestedFilename = response.suggestedFilename();
    m_httpStatusCode = response.httpStatusCode();

    m_expires = response.expires();
    m_maxAge = response.cacheControlMaxAge();
    m_date = isfinite(response.date()) ? response.date() : currentTime();
    m_lastModified = response.lastModified();
    m_lastModifiedHeader = response.httpHeaderField("Last-Modified");
    m_noCache = response.cacheControlContainsNoCache();
    m_mustRevalidate = response.cacheControlContainsMustRevalidate();

    return true;
}

bool HTTPCachedResource::writeFile(String filename, String filepath)
{
    int fd = 0;
    bool result = false;

    if (!m_resourceData) {
        return false;
    }

    String fileFullPath = pathByAppendingComponent(filepath, filename);

    fd = open(fileFullPath.utf8().data(), O_WRONLY | O_CREAT, 0);
    if (fd == -1) {
        return result;
    }

    const char *buf;
    int len, length, written = 0, position = 0;


    for (len = m_resourceData->size(); len > 0;  len -= length) {
        length = m_resourceData->getSomeData(buf, position);
        if (!length) {
            goto error_end;
        }

        for (int writelength = length; writelength > 0; writelength -= written) {
            written = write(fd, (void *)buf, writelength);
            if (written < 0) {
                goto error_end;
            }
        }
        position += length;
    }

    fsync(fd);

    //OS::Report("cache write %s: mimetype: %s\n", m_url.utf8().data(), m_mimeType.utf8().data());

    m_contentLength = m_resourceData->size();
    m_fileName = filename;
    calcResourceSize();

    m_resourceData = 0;

    result = true;
error_end:
    if (fd)
        close(fd);

    return result;
}

bool HTTPCachedResource::readFile(char *out_buf, String filepath)
{
    int fd = 0;
    bool result = false;
    char *buf = out_buf;

    String fileFullPath = pathByAppendingComponent(filepath, m_fileName);

    fd = open(fileFullPath.utf8().data(), O_RDONLY, 0);
    if (fd == -1)
        goto error_end;

    int len, amt;
    for (len = m_contentLength; len > 0; len -= amt) {
        amt = read(fd, buf, len);
        if (amt <= 0)
            break;

        buf += amt; 
    }

    m_used = true;
    result = true;
error_end:
    if (fd != -1)
        close(fd);

    return result;
}

#define ROUNDUP(x, y)   (((x)+((y)-1))/(y)*(y))
#define ROUNDUP_UNIT    4

void HTTPCachedResource::calcResourceSize()
{
    int size = sizeof(long long) * 2
        + sizeof(int) * 3
        + sizeof(double) * 4
        + sizeof(int) + m_url.utf8().length()
        + sizeof(int) + m_mimeType.utf8().length()
        + sizeof(int) + m_textEncodingName.utf8().length()
        + sizeof(int) + m_suggestedFilename.utf8().length()
        + sizeof(int) + m_fileName.utf8().length()
        + sizeof(int) + m_lastModifiedHeader.utf8().length();

    m_resourceSize = ROUNDUP(size, ROUNDUP_UNIT);
}

int writeString(char *buffer, const String &str)
{
    CString cstr = str.utf8();
    int length = cstr.length();

    (*(int*)buffer) = length; buffer += sizeof(int);

    memcpy(buffer, cstr.data(), length);
    return length + sizeof(int);
}

int HTTPCachedResource::serialize(char *buffer)
{
    char *buf = buffer;

    // number field
    (*(long long*)buffer) = m_expectedContentLength; buffer += sizeof(long long);
    (*(long long*)buffer) = m_contentLength; buffer += sizeof(long long);
    (*(int*)buffer) = m_httpStatusCode; buffer += sizeof(int);
    (*(int*)buffer) = m_noCache ? 1 : 0; buffer += sizeof(int);
    (*(int*)buffer) = m_mustRevalidate ? 1 : 0; buffer += sizeof(int);
    (*(double*)buffer) = m_expires; buffer += sizeof(double);
    (*(double*)buffer) = m_maxAge; buffer += sizeof(double);
    (*(double*)buffer) = m_date; buffer += sizeof(double);
    (*(double*)buffer) = m_lastModified; buffer += sizeof(double);
    // string field
    buffer += writeString(buffer, m_url);
    buffer += writeString(buffer, m_mimeType);
    buffer += writeString(buffer, m_textEncodingName);
    buffer += writeString(buffer, m_suggestedFilename);
    buffer += writeString(buffer, m_fileName);
    buffer += writeString(buffer, m_lastModifiedHeader);

    return ROUNDUP(buffer - buf, ROUNDUP_UNIT);
}

int readString(char *buffer, String &str)
{
    int length;
    
    length= (*(int*)buffer); buffer += sizeof(int);
    str = String(buffer, length);

    return length + sizeof(int);
}

int HTTPCachedResource::deserialize(char *buffer)
{
    char *buf = buffer;

    // number field
    m_expectedContentLength = (*(long long*)buffer); buffer += sizeof(long long);
    m_contentLength = (*(long long*)buffer); buffer += sizeof(long long);
    m_httpStatusCode = (*(int*)buffer); buffer += sizeof(int);
    m_noCache = (*(int*)buffer); buffer += sizeof(int);
    m_mustRevalidate = (*(int*)buffer); buffer += sizeof(int);
    m_expires = (*(double*)buffer); buffer += sizeof(double);
    m_maxAge = (*(double*)buffer); buffer += sizeof(double);
    m_date = (*(double*)buffer); buffer += sizeof(double);
    m_lastModified = (*(double*)buffer); buffer += sizeof(double);
    // string field
    buffer += readString(buffer, m_url);
    buffer += readString(buffer, m_mimeType);
    buffer += readString(buffer, m_textEncodingName);
    buffer += readString(buffer, m_suggestedFilename);
    buffer += readString(buffer, m_fileName);
    buffer += readString(buffer, m_lastModifiedHeader);

    calcResourceSize();

    return ROUNDUP(buffer - buf, ROUNDUP_UNIT);
}

size_t HTTPCachedResource::getFileSize()
{ 
    return m_resourceData->size();
}

// HTTPCache

#include <libxml/parser.h>
#include <libxml/tree.h>

#define DEFAULT_FAT_FILENAME        "cache.fat"
#define DEFAULT_FILEPATH            "/"
#define DEFAULT_TOTALCONTENTSSIZE_LIMIT  10 * 1024 * 1024

HTTPCache::HTTPCache()
    : m_fatFileName(DEFAULT_FAT_FILENAME)
    , m_filePath(DEFAULT_FILEPATH)
    , m_writeInterval(60)
    , m_writeLimit(4*1024*1024)
{
    m_disabled = true;
    m_fileNumber = 0;
    m_totalResourceSize = 0;

    m_totalContentsSize = 0;
    m_limitTotalContentsSize = DEFAULT_TOTALCONTENTSSIZE_LIMIT;
}

HTTPCache::~HTTPCache()
{
    forceTerminate();
}

void HTTPCache::forceTerminate()
{
    HTTPCachedResource *resource;

    for (size_t num = 0; num < m_resourceList.size(); num++) {
        resource = m_resourceList[num];
        delete resource;
    }
    m_resourceList.clear();
    m_resources.clear();

    m_totalResourceSize = 0;
    m_totalContentsSize = 0;
}

HTTPCachedResource* HTTPCache::createHTTPCachedResource(KURL &url, RefPtr<SharedBuffer> resourceData, ResourceResponse &response, bool noCache, bool mustRevalidate, double expires, double maxAge)
{
    if (disabled())
        return 0;

    if (m_limitTotalContentsSize < resourceData->size())
        return 0;   // size over

    KURL kurl = removeFragmentIdentifierIfNeeded(url);
    HTTPCachedResource *resource = new HTTPCachedResource(kurl, response);
    if (!resource)
        return 0;
    
    resource->update(noCache, mustRevalidate, expires, maxAge);
    resource->setResourceData(resourceData);
    return resource;
}

bool HTTPCache::addCachedResource(HTTPCachedResource *resource)
{
    if (disabled()) {
        return false;
    }

    //if max age set to 0, do not cache...
    if (cacheOverrideFor(resource) == 0.0) {
        return false;
    }

    m_resources.set(resource->url(), resource);
    m_resourceList.append(resource);
    m_totalResourceSize += resource->resourceSize();
    m_totalContentsSize += resource->contentLength();

    return true;
}

void HTTPCache::updateCachedResource(HTTPCachedResource *resource, RefPtr<SharedBuffer> resourceData, ResourceResponse &response, bool noCache, bool mustRevalidate, double expires, double maxAge)
{
    resource->setResourceResponse(response);
    resource->setResourceData(resourceData);
    resource->update(noCache, mustRevalidate, expires, maxAge);
}

void HTTPCache::removeResource(HTTPCachedResource *resource)
{

    //OS::Report("remove resource: %s\n", resource->fileName().utf8().data());

    m_resources.remove(resource->url());
    for (size_t num = 0; num < m_resourceList.size(); num++) {
        if (m_resourceList[num] == resource) {
            m_resourceList.remove(num);
            break;
        }
    }
    m_totalResourceSize -= resource->resourceSize();
    m_totalContentsSize -= resource->contentLength();
}

void HTTPCache::removeResourceByNumber(int listNumber)
{
    HTTPCachedResource *resource = m_resourceList[listNumber];

    //OS::Report("remove cache file %s", resource->fileName().utf8().data());

    m_resources.remove(resource->url());
    m_resourceList.remove(listNumber);
    m_totalResourceSize -= resource->resourceSize();
    m_totalContentsSize -= resource->contentLength();

    String fileFullPath = pathByAppendingComponent(getFilePath(), resource->fileName());
    unlink(fileFullPath.utf8().data());

    delete resource;
}

void HTTPCache::remove(HTTPCachedResource *resource)
{
    removeResource(resource);

    String fileFullPath = pathByAppendingComponent(getFilePath(), resource->fileName());

    //OS::Report("remove file %s", fileFullPath.utf8().data());
    unlink(fileFullPath.utf8().data());

    delete resource;
}

void HTTPCache::detach(HTTPCachedResource *resource)
{
    removeResource(resource);

	// since we quite often reach the max size of the cache, it is entirely possible (and very common) that
	// this resource ends up getting removed from the write cache list altogether. so we remove its file because we don't want it dangling
    String fileFullPath = pathByAppendingComponent(getFilePath(), resource->fileName());
	//OS::Report("remove file %s", fileFullPath.utf8().data());
    unlink(fileFullPath.utf8().data());
}

void HTTPCache::removeAll()
{
    forceTerminate();

    m_fileNumber = 0;

    DIR *dir = opendir(getFilePath().utf8().data());
    if (!dir) {
        return;
    }

    dirent *ent;

    while (ent = readdir(dir)) {
        
        String fileFullPath = pathByAppendingComponent(getFilePath(), ent->d_name);
        unlink(fileFullPath.utf8().data());
    }

    closedir(dir);

}

void HTTPCache::removeNotInFATFile()
{
    DIR *dir = opendir(getFilePath().utf8().data());
    if (!dir) {
        return;
    }

    dirent *ent;

    HashMap<String, int> resourceSet;
	int count = 0;
	for (Vector<HTTPCachedResource *>::iterator iter = m_resourceList.begin(); iter != m_resourceList.end(); iter++, count++) {
		resourceSet.add((*iter)->fileName(), count);
	}

    while (ent = readdir(dir)) {
        bool found = resourceSet.contains(ent->d_name); 

        if (!strcmp(ent->d_name, DEFAULT_FAT_FILENAME)) {
            continue;
        }

        if (!found) {
            String fileFullPath = pathByAppendingComponent(getFilePath(), ent->d_name);
            unlink(fileFullPath.utf8().data());
        }
    }

    closedir(dir); 
}

void HTTPCache::setLimitSize(long long limit)
{
    m_limitTotalContentsSize = limit;

    if (limit < m_totalContentsSize) {
        purgeBySize(m_totalContentsSize - limit);
        writeFATFile();
    }
}

void HTTPCache::setFilePath(char *path)
{
    if (!path || !strlen(path))
        return;

    // TODO: If path is same as the default one, readFATFile will never be called even if HTTPCache is enabled...
    if (getFilePath() == path)
        return;

    m_filePath = String(path);
    makeAllDirectories(m_filePath);

    readFATFile();
}

KURL HTTPCache::removeFragmentIdentifierIfNeeded(const KURL& originalURL)
{
    if (!originalURL.hasFragmentIdentifier())
        return originalURL;
    // Strip away fragment identifier from HTTP URLs.
    // Data URLs must be unmodified. For file and custom URLs clients may expect resources 
    // to be unique even when they differ by the fragment identifier only.
    if (!originalURL.protocolInHTTPFamily())
        return originalURL;
    KURL url = originalURL;
    url.removeFragmentIdentifier();
    return url;
}

HTTPCachedResource* HTTPCache::resourceForURL(const KURL& resourceURL)
{
    KURL url = removeFragmentIdentifierIfNeeded(resourceURL);
    HTTPCachedResource* resource = m_resources.get(url);

    return resource;
}

bool HTTPCache::equalHTTPCachedResourceURL(HTTPCachedResource *resource, KURL& resourceURL)
{
    KURL url = removeFragmentIdentifierIfNeeded(resourceURL);
    if (equalIgnoringCase(resource->url(), url.string())) {
        return true;
    }

    return false;
}

String HTTPCache::makeFileName()
{
    String fileName;

    fileName = String::format("%08d.dcf", m_fileNumber);
    m_fileNumber++;

    return fileName;
}

long long HTTPCache::purgeBySizeInternal(long long size)
{
    HTTPCachedResource *resource;
    long long purgedSize = 0;

    for (size_t num = 0; num < m_resourceList.size(); ) {
        resource = m_resourceList[num];
        if (resource->isUsed())
            num++;
        else {
            purgedSize += resource->contentLength();
            removeResourceByNumber(num);
            if (purgedSize > size)
                break;
        }
    }
    return purgedSize;
}

long long HTTPCache::purgeBySize(long long size)
{
    long long purgedSize = purgeBySizeInternal(size);
    if (purgedSize < size) {
        for (size_t num = 0; num < m_resourceList.size(); num++) {
            m_resourceList[num]->setUsed(false);
        }

        purgedSize +=purgeBySizeInternal(size - purgedSize);
    }
    return purgedSize;
}

bool HTTPCache::write(HTTPCachedResource *resource)
{ 
    //if max age set to 0, do not cache...
    if (cacheOverrideFor(resource) == 0.0) {
        return false;
    }

    long long contentLength = resource->contentLength();
    if (m_totalContentsSize + contentLength > m_limitTotalContentsSize) {
        purgeBySize(contentLength);
    }

    if (!resource->writeFile(makeFileName(), getFilePath())) {
        return false;
    }
    
    m_resources.add(resource->url(), resource);
    m_resourceList.append(resource);
    m_totalResourceSize += resource->resourceSize();
    m_totalContentsSize += resource->contentLength();

    return true;
}

bool HTTPCache::read(HTTPCachedResource *resource, char *buf)
{
    return resource->readFile(buf, getFilePath());
}

void HTTPCache::serializeFATData(char *buffer)
{
	int write;
    for (size_t num = 0; num < m_resourceList.size(); num++) {
        buffer += m_resourceList[num]->serialize(buffer);
    }
}

bool HTTPCache::deserializeFATData(char *buffer, int length)
{
    int read = 0;
    HTTPCachedResource *resource;

    for (; length > FAT_HEADER_SIZE; length -= read) {
        resource = new HTTPCachedResource();
        if (!resource) {
            return false;
        }

        read = resource->deserialize(buffer);
        buffer += read;

        //if (!resource->fileName().isEmpty()) {
            addCachedResource(resource);
        //}
    }    
    return true;
}

#define CACHEFAT_FORMAT_VERSION     1  // Number of int. Increment this if you changed the content format in the fat file.
#define CACHE_READBUFFER_SIZE       8192

#define MD5_DIGESTSIZE 16

void HTTPCache::writeDigest(unsigned char *data, int length)
{
    MD5 md5;
    md5.addBytes(data + MD5_DIGESTSIZE, length - MD5_DIGESTSIZE);
    Vector<uint8_t, 16> digest;
    md5.checksum(digest);
    memcpy(data, digest.data(), MD5_DIGESTSIZE);
}

bool HTTPCache::writeFATFile()
{
    char *buf, *buffer;
    bool result = false;

    int totalSize = 0;
    int headerSize = 0;

    headerSize = sizeof(int) * 2 + MD5_DIGESTSIZE;
    totalSize = headerSize + m_totalResourceSize;

    //OS::Report("writing FAT file: totalSize: %d\n", totalSize);

    buf = (char*)fastZeroedMalloc(totalSize);
    if (!buf) {
        return false;
    }

    buffer = buf + MD5_DIGESTSIZE;
    (*(int*)buffer) = CACHEFAT_FORMAT_VERSION; buffer += sizeof(int);
    (*(int*)buffer) = m_fileNumber; buffer += sizeof(int);

    serializeFATData(buffer);

    writeDigest((unsigned char*)buf, totalSize);

    String fileFullPath = pathByAppendingComponent(getFilePath(), m_fatFileName);

    int fd, len, written = 0;
    fd = open(fileFullPath.utf8().data(), O_WRONLY | O_CREAT, 0);
    if (fd == -1) {
        goto error_end;
    }

    buffer = buf;
    for (len = totalSize; len > 0;  len -= written) {
        written = ::write(fd, buffer, len);
        if (written <= 0) {
            break;
        }
        buffer += written;
    }

    if (len > 0) {
        goto error_end;
    }

    fsync(fd);

    result = true;
error_end:
    if (fd != -1) {
        close(fd);
    }

    fastFree(buf);
    return result;
}

bool HTTPCache::verifyDigest(unsigned char *data, int length)
{
    if (length <= MD5_DIGESTSIZE) {
        return false;
    }

    bool result = false;
    MD5 md5;
    md5.addBytes(data + MD5_DIGESTSIZE, length - MD5_DIGESTSIZE);
    Vector<uint8_t, 16> digest;
    md5.checksum(digest);

    if (memcmp(data, digest.data(), MD5_DIGESTSIZE) == 0) {
        result = true;
    }

    return result;
}

bool HTTPCache::readFATFile()
{
    int fd;
    bool result = false, remove = false;
    char *buf, *buffer;
    int format_version = 0;
    int len = 0, amt = 0;

    buf = (char*)fastZeroedMalloc(CACHE_READBUFFER_SIZE);
    if (!buf) {
        return false;
    }

    String fileFullPath = pathByAppendingComponent(getFilePath(), m_fatFileName);

    fd = open(fileFullPath.utf8().data(), O_RDONLY, 0);
    if (fd == -1) {
        goto error_end;
    }

    for (;;) {
        amt = ::read(fd, buf + len, CACHE_READBUFFER_SIZE);
        if (amt < 0) {
            goto error_end;
        }

        if (amt == 0 ) {
            break;
        }

        len += amt;
        buf = (char*)fastRealloc(buf, len + CACHE_READBUFFER_SIZE);
        if (!buf) {
            goto error_end;
        }
    }

    buffer = buf;

    if (!verifyDigest((unsigned char*)buffer, len)) {
        //OS::Report("FAT file read error: MD5 check");
        remove = true;
        goto error_end;
    }

    buffer += MD5_DIGESTSIZE;

    format_version = (*(int*)buffer); buffer += sizeof(int);
    if (format_version != CACHEFAT_FORMAT_VERSION) {
        //OS::Report("FAT file format is old : %d", format_version);
        remove = true;
        goto error_end;
    }

    m_fileNumber = (*(int*)buffer); buffer += sizeof(int);
    len -= sizeof(int) + MD5_DIGESTSIZE;

    if (deserializeFATData(buffer, len)) {
        result = true;
    } else {
        remove = true;
    }

    removeNotInFATFile();

error_end:
    if (fd != -1) {
        close(fd);
    }

    if (remove) {
        removeAll();
    }

    fastFree(buf);
    return result;
}

int HTTPCache::getFATFileSize()
{
    int totalSize = 0;
    int headerSize = 0;

    headerSize = sizeof(int) * 2 + MD5_DIGESTSIZE;
    totalSize = headerSize + m_totalResourceSize;

    return totalSize;
}

double HTTPCache::getWriteInterval()
{
    return m_writeInterval;
}

size_t HTTPCache::getWriteLimit()
{
    return m_writeLimit;
}

double HTTPCache::cacheOverrideFor(HTTPCachedResource *resource)
{
    double maxage = -1.0;
    String domain;

    if (m_cacheOverrides.contains(resource->mimeType())) {
        maxage = m_cacheOverrides.get(resource->mimeType())->find(resource->url());
        //CRANK RJD:  Check for domain override.
        if(maxage == -1.0) {
        	domain = getDomain(resource->url());
        	maxage = m_cacheOverrides.get(resource->mimeType())->find(domain);
        }
        //OS::Report("override result for url: %s mimeType: %s is %f\n", resource->url().utf8().data(), resource->mimeType().utf8().data(), maxage);
    }

    if (maxage == -1.0 && m_cacheOverrides.contains(String("all"))) {
        maxage = m_cacheOverrides.get("all")->find(resource->url());
        //CRANK RJD:  Check for domain override.
        if(maxage == -1.0) {
        	domain = getDomain(resource->url());
           	maxage = m_cacheOverrides.get("all")->find(domain);
         }
        //OS::Report("override result for url: %s all is %f\n",resource->url().utf8().data(),  resource->mimeType().utf8().data(), maxage); 
    }

    return maxage;
}

String &HTTPCache::getFilePath()
{
    return m_filePath;
}

String HTTPCache::getDomain(String url) {
	String dot = String(".");
	String domain;
	int start = 0;
	int end = url.length();
	int len;



	if (url.startsWith("http://")  == true) start = 7;
	if (url.startsWith("https://") == true) start = 8;
	if (url.find("/", start) != notFound) end = url.find('/', start);
	if (url.find(":", start) != notFound) end = url.find(':', start);


	int lastDot = url.reverseFind(dot, end);

	//RJD CRANK:  No dots found...just return what we have.
	if (lastDot == notFound) {
		domain = url.substring(start, end);
		return domain;
	}
	int secondLastDot = url.reverseFind(dot, lastDot - 1);

	// RJD CRANK:  Someone put something like http://google.com in the file, or google.com in the config
	if(secondLastDot == notFound) {
		secondLastDot = start;
		len = end - secondLastDot;
	} else {
		secondLastDot++;
	}

	 domain = url.substring(secondLastDot,  end - secondLastDot);
	return domain;
}

void HTTPCache::parseConfigFile()
{
    xmlDocPtr doc;

    const char *config = shim_GetConfigXMLFile();
    int configSize = shim_GetConfigXMLSize();

    if (configSize) {
        doc = xmlReadMemory(config, configSize, NULL, NULL, 0);
    }
    else {
        doc = xmlReadFile(config, NULL, 0);
    }
    if (doc == NULL) {
        //OS::Report("warning: could not parse cache configuration file: using default values");
        return;
    }

    xmlNode *root = NULL, *cur = NULL;
    root = xmlDocGetRootElement(doc);

    //get root of cache configuration
    for (cur = root->children; cur; cur = cur->next) {

        if (xmlStrEqual(cur->name, (xmlChar *)"options")) {
            //found options, look for cache config in its children
            cur = cur->children;
            continue;
        }

        if (xmlStrEqual(cur->name, (xmlChar *)"cacheConfig")) {
            //found cache config, we can stop searching
            root = cur; 
            break;
        }
    }

    if (root == xmlDocGetRootElement(doc)) {
        //OS::Report("warning: configuration file has no cache section: using default values");
        return; 
    }

    for (cur = root->children; cur; cur = cur->next) {
        if (cur->type == XML_ELEMENT_NODE) {
            xmlNode *child = cur->children;

            //xmlChar is typedef'd to be unsigned char, so we end up
            //doing a lot of casts back to char below: 
            if (child) {

                if (xmlStrEqual(cur->name, (xmlChar *)"nandCacheEnabled")) {
                    if (child->content) {
                        if (xmlStrEqual(child->content, (xmlChar *)"1")) {
                            //OS::Report("enabling cache...\n");
                            m_disabled = false;
                        } else {
                            //OS::Report("disabling cache...\n"); 
                        }
                    }
                } else if (xmlStrEqual(cur->name, (xmlChar *)"savePath")) {
                    if (child->content) {
                        setFilePath((char *)child->content);
                        //OS::Report("setting path to: %s\n", child->content); 
                    }
                } else if (xmlStrEqual(cur->name, (xmlChar *)"maxSize")) {
                    if (child->content) {
                        m_limitTotalContentsSize = atoi((char *)child->content); 
                        //OS::Report("setting content size limit to: %lld\n", m_limitTotalContentsSize); 
                    }
                } else if (xmlStrEqual(cur->name, (xmlChar *)"writeInterval")) {
                    if (child->content) { 
                        m_writeInterval = atof((char *)child->content);
                        //enforce lotcheck guidelines
                        if (m_writeInterval < 60.0) {
                            m_writeInterval = 60.0;
                        }
                        //OS::Report("setting write interval to: %f\n", m_writeInterval);
                    }
                } else if (xmlStrEqual(cur->name, (xmlChar *)"writeSize")) {
                    m_writeLimit = atoi((char *)child->content);

                    //enforce lotcheck guidelines
                    if (m_writeLimit > 8*1024*1024) {
                        m_writeLimit = 8*1024*1024;
                    }
                    //OS::Report("setting write limit to: %d\n", m_writeLimit);
                } else if (xmlStrEqual(cur->name, (xmlChar *)"cacheOverride")) {
                    xmlChar *url = xmlGetProp(cur, (xmlChar *)"url");
                    xmlChar *allowSubdomains = xmlGetProp(cur, (xmlChar *)"subdomains");
                    String key = String((char *)url);
                    if(allowSubdomains != NULL) {
                    	if(strcmp((char*)allowSubdomains, "true") == 0) {
                    		String tmpkey = getDomain(key);
                    		key = tmpkey;
                    	}
                    }
                    //OS::Report("url=%s\n", url);
                    for (xmlNode *node = cur->children; node; node = node->next) {
                        if (xmlStrEqual(node->name, (xmlChar *)"maxage")) {
                            if (node->children && node->children->content) {
                                xmlChar *prop = xmlGetProp(node, (xmlChar *)"mime-type");

                                String mimeType((char *)prop);
                                double maxAge = atof((char *)node->children->content);

                                if (m_cacheOverrides.contains(mimeType)) {
                                    m_cacheOverrides.get(mimeType)->insert(key, maxAge);
                                } else {
                                    Trie<double> *t = new Trie<double>(-1);
                                    t->insert(key, maxAge);
                                    m_cacheOverrides.set(mimeType, t); 
                                }

                                //OS::Report("maxtime: %s\n", node->children->content);
                                //OS::Report("mime-type: %s\n", prop? prop: "all");
                            }
                        }
                    }

                } else {
                    //OS::Report("unrecognized node in cache config.xml: %s\n", cur->name);
                    handleUnrecognizedConfigNode(cur);
                }

            }
        }
    }

    xmlFreeDoc(doc);

    return;
}

HTTPSCache::HTTPSCache()
    : HTTPCache()
    , m_filePath(DEFAULT_FILEPATH)
{
    m_disabled = true; 
}

HTTPSCache::~HTTPSCache()
{

}

void HTTPSCache::handleUnrecognizedConfigNode(xmlNode *node)
{
    xmlNode *child = node->children;

    //xmlChar is typedef'd to be unsigned char, so we end up
    //doing a lot of casts back to char below: 
    if (child) {

        if (xmlStrEqual(node->name, (xmlChar *)"nandCacheEnabledHTTPS")) {
            if (child->content) {
                if (xmlStrEqual(child->content, (xmlChar *)"1")) {
                    //OS::Report("enabling https cache...\n");
                    m_disabled = false;
                } else {
                    //OS::Report("disabling https cache...\n"); 
                }
            } 
        } else if (xmlStrEqual(node->name, (xmlChar *)"savePathHTTPS")) {
            if (child->content) {
                StringBuilder sb;
                //sb.append(shim_GetFSSaveAccountDir());
                sb.append((char *)child->content);
                m_filePath = sb.toString(); 

                makeAllDirectories(m_filePath);
                readFATFile();
                //OS::Report("setting https path to: %s\n", m_filePath.utf8().data()); 
            }
        } 
    } 
}

} // namespace
