/*
 * Copyright (c) 2011,2012 ACCESS CO., LTD. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 * 
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 */

#ifndef HTTPCache_h
#define HTTPCache_h

#include "PlatformString.h"
#include "SharedBuffer.h"
#include "ResourceResponse.h"
#include "CurrentTime.h"
#include <wtf/HashMap.h>
#include <wtf/ListHashSet.h>
#include <wtf/Vector.h>
#include <wtf/text/StringHash.h>

#include <Trie.h>

namespace WebCore {

class SharedBuffer;
class KURL;
class ResourceResponse;

class _xmlNode;

class HTTPCachedResource {
public:
    HTTPCachedResource();
    HTTPCachedResource(KURL &url, ResourceResponse &response);
    ~HTTPCachedResource();

    bool setResourceResponse(ResourceResponse &response);
    double freshnessLifetime();
    bool isExpired();
    bool needRevalidate(double override);
    void update(bool noCache, bool mustRevalidate, double expires, double maxAge);

    const String& url() { return m_url; }
    const String& mimeType() { return m_mimeType; }
    long long expectedContentLength() { return m_expectedContentLength; }
    const String& suggestedFilename() { return m_suggestedFilename; }
    const String& textEncodingName() { return m_textEncodingName; }
    long long contentLength() { return m_contentLength; }
    const String& fileName() { return m_fileName; }
    int httpStatusCode() { return m_httpStatusCode; }
    double expires() { return m_expires; }
    const String& lastModifiedHeader() { return m_lastModifiedHeader; }

    bool writeFile(String filename, String filepath);
    bool readFile(char *buf, String filepath);

    void calcResourceSize();
    int resourceSize() { return m_resourceSize; }

    int serialize(char *buffer);
    int deserialize(char *buffer);

    void setResourceData(RefPtr<SharedBuffer> resourceData) { m_resourceData = resourceData, m_contentLength = resourceData->size(); }

    bool isUsed() { return m_used; }
    void setUsed(bool used) { m_used = used; }

    size_t getFileSize(); 

private:
    String m_url;
    String m_mimeType;
    long long m_expectedContentLength;
    String m_textEncodingName;
    String m_suggestedFilename;
    int m_httpStatusCode;
    bool m_noCache;
    bool m_mustRevalidate;
    double m_expires;
    double m_maxAge;
    double m_date;
    double m_lastModified;
    String m_lastModifiedHeader;
    String m_fileName;

    RefPtr<SharedBuffer> m_resourceData;
    long long m_contentLength;
    int m_resourceSize;
    bool m_used;
};

typedef HashMap<String, HTTPCachedResource*> HTTPCachedResourceMap;
typedef HashMap<String, HTTPCachedResource*>::iterator HTTPCachedResourceMapIterator;

class HTTPCache {
public:
    HTTPCache();
    virtual ~HTTPCache();
    void forceTerminate();

    HTTPCachedResource* createHTTPCachedResource(KURL &url, RefPtr<SharedBuffer> resourceData, ResourceResponse &response, bool noCache, bool mustRevalidate, double expires, double maxAge);
    bool addCachedResource(HTTPCachedResource *resource);
    void updateCachedResource(HTTPCachedResource *resource, RefPtr<SharedBuffer> resourceData, ResourceResponse &response, bool noCache, bool mustRevalidate, double expires, double maxAge);
    void removeResource(HTTPCachedResource *resource);
    void removeResourceByNumber(int listNumber);
    void remove(HTTPCachedResource *resource);
    void detach(HTTPCachedResource *resource);
    void removeAll();
    void removeNotInFATFile();

    virtual bool disabled() const { return m_disabled; }
    void setLimitSize(long long limit);
    void setFilePath(char* path);

    KURL removeFragmentIdentifierIfNeeded(const KURL& originalURL);
    HTTPCachedResource* resourceForURL(const KURL& resourceURL);
    bool equalHTTPCachedResourceURL(HTTPCachedResource *resource, KURL& resourceURL);

    String makeFileName();
    long long purgeBySizeInternal(long long size);
    long long purgeBySize(long long size);
    bool write(HTTPCachedResource *resource);
    bool read(HTTPCachedResource *resource, char *buf);

    void serializeFATData(char *buffer);
    bool deserializeFATData(char *buffer, int length);
    bool writeFATFile();
    bool readFATFile();
    int getFATFileSize();

    void writeDigest(unsigned char *data, int length);
    bool verifyDigest(unsigned char *data, int length);

    double getWriteInterval();
    size_t getWriteLimit();

    double cacheOverrideFor(HTTPCachedResource *resource);

    virtual String &getFilePath();

    void parseConfigFile();

protected:

    virtual void handleUnrecognizedConfigNode(_xmlNode *node) {};

private:

	String getDomain(String url);

    HashMap<String, Trie<double> *> m_cacheOverrides;

    bool m_disabled;
    HTTPCachedResourceMap m_resources;
    Vector<HTTPCachedResource*> m_resourceList;

    int m_fileNumber;
    int m_totalResourceSize;

    String m_fatFileName;
    String m_filePath;

    long long m_totalContentsSize;
    long long m_limitTotalContentsSize;

    double m_writeInterval;
    size_t m_writeLimit;
};

class HTTPSCache : public HTTPCache {

public:

    HTTPSCache();
    virtual ~HTTPSCache();
    virtual String &getFilePath() { return m_filePath; }
    virtual bool disabled() const { return HTTPCache::disabled() || m_disabled; }

protected:

    virtual void handleUnrecognizedConfigNode(_xmlNode *node);

private:

    String m_filePath;
    bool m_disabled;
};

} // namespace


#endif //HTTPCache_h
