/*
 * Copyright (C) 2006, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2008 INdT - Instituto Nokia de Tecnologia
 * Copyright (C) 2009-2010 ProFUSION embedded systems
 * Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "PopupMenuWiiU.h"

#include "Chrome.h"
#include "ChromeClientWiiU.h"
#include "Frame.h"
#include "FrameView.h"
#include "NotImplemented.h"
#include "Page.h"
#include "PopupMenuClient.h"
#include "TextRun.h"

static const int maxPopupHeight = 360;
static const int popupWindowBorderWidth = 1;

namespace WebCore {

PopupMenuWiiU::PopupMenuWiiU(PopupMenuClient* client)
    : m_popupClient(client)
    , m_view(0)
	, m_scrollOffset(0)
	, m_scrollbarCapturingMouse(false)
{
}

PopupMenuWiiU::~PopupMenuWiiU()
{
    // Tell client to destroy data related to this popup since this object is
    // going away.
    if (m_view)
        hide();
    if (m_scrollbar)
        m_scrollbar->setParent(0);
}

const int separatorPadding = 4;
const int separatorHeight = 1;
void PopupMenuWiiU::paint(GraphicsContext *ctx) 
{

    ASSERT(m_popupClient);
 
    int itemCount = m_popupClient->listSize();
    PopupMenuStyle itemStyle = m_popupClient->itemStyle(0); 
    const Font &itemFont = itemStyle.font();

    ctx->save();
    ctx->clip(m_drawRect);
    ctx->translate(m_drawRect.x(), m_drawRect.y());

    // listRect is the damageRect translated into the coordinates of the entire menu list (which is itemCount * m_itemHeight pixels tall)

    IntRect listRect = IntRect(0, 0, m_clientRect.width(), m_clientRect.height());
    listRect.move(IntSize(0, m_scrollOffset * m_itemHeight));

    for (int y = listRect.y(); y < listRect.maxY(); y += m_itemHeight) {
    	int index = y / m_itemHeight;

    	Color optionBackgroundColor, optionTextColor;
    	itemStyle = m_popupClient->itemStyle(index);

    	optionBackgroundColor = itemStyle.backgroundColor();
    	optionTextColor = itemStyle.foregroundColor();

        // itemRect is in client coordinates
        IntRect itemRect(0, (index - m_scrollOffset) * m_itemHeight,  m_drawRect.width(), m_itemHeight);
        // Draw the background for this menu item
        if (itemStyle.isVisible())
        	ctx->fillRect(itemRect, optionBackgroundColor, ColorSpaceDeviceRGB);

        if (m_popupClient->itemIsSeparator(index)) {
        	IntRect separatorRect(itemRect.x() + separatorPadding, itemRect.y() + (itemRect.height() - separatorHeight) / 2, itemRect.width() - 2 * separatorPadding, separatorHeight);
        	ctx->fillRect(separatorRect, optionTextColor, ColorSpaceDeviceRGB);
        	continue;
        }

        String itemText = m_popupClient->itemText(index);

        unsigned length = itemText.length();
        const UChar* string = itemText.characters();
        TextDirection direction = (itemText.defaultWritingDirection() == WTF::Unicode::RightToLeft) ? RTL : LTR;
        TextRun textRun(string, length, false, 0, 0, TextRun::AllowTrailingExpansion, direction);

        ctx->setFillColor(optionTextColor, ColorSpaceDeviceRGB);

        // Draw the item text
        if (itemStyle.isVisible()) {
        	int textX = itemRect.x() + m_popupClient->clientPaddingLeft() - m_popupClient->clientInsetLeft();
        	if (textX < itemRect.x()) {
        		textX = itemRect.x();
        	}
        	int textY = itemRect.y() + itemFont.fontMetrics().ascent() + (itemRect.height() - itemFont.fontMetrics().height()) / 2;
        	ctx->drawBidiText(itemFont, textRun, IntPoint(textX, textY));
        }
    }

    //Draw border lines
    ctx->translate(-m_drawRect.x(), -m_drawRect.y());
    ctx->setStrokeColor(itemStyle.foregroundColor(), ColorSpaceDeviceRGB);
    ctx->drawLine(IntPoint(m_drawRect.x(), m_drawRect.y()), IntPoint(m_drawRect.x(), m_drawRect.y() + m_drawRect.height()-1));
    ctx->drawLine(IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y()), IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y() + m_drawRect.height()-1));
    if(m_drawAbove) {
    	ctx->drawLine(IntPoint(m_drawRect.x(), m_drawRect.y()), IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y()));
    } else {
    	ctx->drawLine(IntPoint(m_drawRect.x(), m_drawRect.y() + m_drawRect.height()-1), IntPoint(m_drawRect.x() + m_drawRect.width()-1, m_drawRect.y() + m_drawRect.height()-1));
    }

    ctx->restore();

    if (m_scrollbar) {
    	m_scrollbar->paint(ctx, m_drawRect);
    }

}

void PopupMenuWiiU::calculatePositionAndSize(const IntRect& rect, FrameView* view) 
{
    int itemCount = m_popupClient->listSize();
    m_itemHeight = m_popupClient->menuStyle().font().fontMetrics().height() + 1;
    int popupHeight = m_itemHeight * (itemCount);
    if (popupHeight > maxPopupHeight) {
        popupHeight = maxPopupHeight;
    }
    // The popup should show an integral number of items (i.e. no partial items should be visible)
    popupHeight -= popupHeight % m_itemHeight;
    IntRect transRect = rect;
    transRect.move( -m_view->scrollPosition().x(), -m_view->scrollPosition().y());

    FloatRect windowRect = transRect;
    m_drawAbove = false;

    Frame *frame = view->frame();
    if(frame != NULL) {
    	Page *page = frame->page();
    	if(page != NULL) {
   		Chrome *chrome = page->chrome();
    		if(chrome != NULL) {
    			windowRect = chrome->windowRect();
    		}
    	}
    }

    m_clientRect.setX(0);
    m_clientRect.setY(0);
    m_clientRect.setWidth(transRect.width() - (popupWindowBorderWidth * 2));
    m_clientRect.setHeight(popupHeight);

    // For now, assume we are keepig it the same width as the text portion of the combobox
    m_drawRect.setX(transRect.x());
    if((transRect.height() != windowRect.height()) && ((transRect.y() + transRect.height() + popupHeight - 1) > windowRect.height())) {
    	m_drawRect.setY(transRect.y() - popupHeight);
    	m_drawAbove = true;
    } else {
    	m_drawRect.setY(transRect.y() + transRect.height());
    }
    m_drawRect.setWidth(transRect.width());
    m_drawRect.setHeight(popupHeight + popupWindowBorderWidth);

}

void PopupMenuWiiU::show(const IntRect& rect, FrameView* view, int index)
{
    ASSERT(m_popupClient);
    m_view = view;

    calculatePositionAndSize(rect, view);

    if (!m_scrollbar && visibleItems() < client()->listSize()) {
    	// We need a scroll bar
    	m_scrollbar = client()->createScrollbar(this, VerticalScrollbar, SmallScrollbar);
        m_scrollbar->setFrameRect(IntRect(m_drawRect.x() + m_drawRect.width() - m_scrollbar->width() - 1, m_drawRect.y() + 1, m_scrollbar->width(), m_drawRect.height() - 2));
        int visibleItems = this->visibleItems();
        m_scrollbar->setEnabled(visibleItems < client()->listSize());
        m_scrollbar->setSteps(1, max(1, visibleItems - 1));
        m_scrollbar->setProportion(visibleItems, client()->listSize());
    	m_scrollbar->styleChanged();
    }

    ChromeClientWiiU* chromeClient = static_cast<ChromeClientWiiU*>(view->frame()->page()->chrome()->client());
    ASSERT(chromeClient);
    chromeClient->showPopup(this, m_drawRect);
    m_view->hostWindow()->invalidateContentsAndWindow(m_drawRect, false);
}

void PopupMenuWiiU::hide()
{
    if (m_view && m_popupClient) { 
        m_popupClient->popupDidHide();

        ChromeClientWiiU* chromeClient = static_cast<ChromeClientWiiU*>(m_view->frame()->page()->chrome()->client());
        ASSERT(chromeClient);
        chromeClient->hidePopup();
        m_view->hostWindow()->invalidateContentsAndWindow(IntRect(m_drawRect.x() - 1, m_drawRect.y(), m_drawRect.width() + 2, m_drawRect.height()), false);
    }
}

void PopupMenuWiiU::processEvent(PlatformMouseEvent ev)
{
    //should always be true
    if (!m_popupClient) {
    	return;
    }

    IntPoint mousePoint = IntPoint(ev.x(), ev.y());

    if(m_scrollbar) {
    	IntRect scrollBarRect = m_scrollbar->frameRect();
    	if (scrollBarRect.contains(mousePoint) || m_scrollbarCapturingMouse) {
    		mousePoint.move(-scrollBarRect.x(), -scrollBarRect.y());
    		PlatformMouseEvent event(ev);
    		event.setPosition(mousePoint.x(), mousePoint.y());
    		if(ev.eventType() == MouseEventMoved ) {
    			if(m_scrollbarCapturingMouse == true) {
    				m_scrollbar->mouseMoved(event);
    			}
    		} else if(ev.eventType() == MouseEventPressed) {
    			m_scrollbar->mouseDown(event);
    			m_scrollbarCapturingMouse = true;
    		} else if(ev.eventType() == MouseEventReleased) {
    			m_scrollbarCapturingMouse = false;
    			m_scrollbar->mouseUp();
    		}
    		m_view->hostWindow()->invalidateContentsAndWindow(m_drawRect, false);
    		return;
    	}
    }

    if(m_scrollbarCapturingMouse == true) {
    	return;
    }
    if(ev.eventType() != MouseEventPressed) {
    	return;
    }
    int new_index = m_popupClient->selectedIndex();
    if (m_drawRect.contains(ev.x(), ev.y())) {
    	int potential_new_index = (ev.y() + (m_scrollOffset * m_itemHeight) - m_drawRect.y()) / m_itemHeight;

    	//this should be fine due to the contains() call above, but why not be safe...
    	if (potential_new_index >= 0 && potential_new_index < m_popupClient->listSize()) {
    		new_index = potential_new_index;
    	}
    }

    m_popupClient->valueChanged(new_index);
    hide();
}

void PopupMenuWiiU::updateFromElement()
{
    m_popupClient->setTextFromItem(m_popupClient->selectedIndex());
}

void PopupMenuWiiU::disconnectClient()
{
    if (m_view)
        hide();
    m_popupClient = 0;
}

void PopupMenuWiiU::setScrollOffset(const IntPoint& offset)
{
    scrollTo(offset.y());
}

void PopupMenuWiiU::scrollTo(int offset)
{
    ASSERT(m_scrollbar);

    if (m_scrollOffset == offset)
        return;

    m_scrollOffset = offset;
}

int PopupMenuWiiU::scrollSize(ScrollbarOrientation orientation) const
{
    return ((orientation == VerticalScrollbar) && m_scrollbar) ? (m_scrollbar->totalSize() - m_scrollbar->visibleSize()) : 0;
}

int PopupMenuWiiU::scrollPosition(Scrollbar*) const
{
    return m_scrollOffset;
}

int PopupMenuWiiU::visibleItems() const
{
    return clientRect().height() / m_itemHeight;
}

IntRect PopupMenuWiiU::clientRect() const
{
    return m_clientRect;
}


void PopupMenuWiiU::invalidateScrollbarRect(Scrollbar* scrollbar, const IntRect& rect)
{
    IntRect scrollRect = rect;
    scrollRect.move(scrollbar->x(), scrollbar->y());
}

}

