#ifndef ERROR_H
#define ERROR_H

#ifdef PF_UNIXBLD
#	include <unicode/utypes.h>
#endif
#include <string>
#include <map>

#ifdef _WIIU_
#include <shim.h>
#include <thread_ctrl.h>
#endif

#ifdef WIN32
/* FIXME: CLEANUP -- because mingw-gcc has its own headers for e.g. <sys/...>,
    we have to avoid adding 'libshim/include' as a -I path, and hence need to
    spell out fuller #include paths for headers that *are* within libshim.
    There must be a better way to organize this.
   -TS-NSA 2012-10-08 */
#include <libshim/include/pthread.h>
#else
#include <pthread.h>
#endif

namespace LibPlatform {
	class Error {
	public:
		/**
		This structure defines the possible states of the error ui
		*/
		enum State {
			SHOW = 0,
			SHOW_NO_PAUSE,
			HIDE,
			HIDDEN
		};

		/**
		This structure defines the possible states of the error ui
		*/
		enum ScreenType {
			DISPLAY_TV = 0,
			DISPLAY_GamePad_0,
			DISPLAY_GamePad_1,
			DISPLAY_ALL
		};

		/**
		APW - 9/18: I moved the ErrorCode enum to another file because it is huge
		**/
		#include "PlatformErrorCodes.h"


		/**
		Visibility types: Which class of errors are shown to the user
		*/
		enum Visibility {
			CORE = 0,
			APPLICATION,
			USER,
		};

		/**
			Which button was pressed on the window.
			Center(or one button) doubles for left.
		*/
		enum ErrorButton{
			LEFT_BTN = 0,
			RIGHT_BTN,
			NO_BTN,
		};

		/**
			Data passed into the callback
		*/
		struct ErrorCallbackData
		{
			State state;
			ErrorButton buttonPressed;
		};

		struct CustomErrorData
		{
			std::wstring bodyText;
			std::wstring btn1Text;
			std::wstring btn2Text;
			CustomErrorData()
			{
			}
			CustomErrorData(std::wstring body, std::wstring btn1, std::wstring btn2)
				:bodyText(body),
				btn1Text(btn1),
				btn2Text(btn2)
			{}
		};

		typedef void (*StateChangeCallback)(ErrorCallbackData);

		// Data queued for reporting errors
		struct ReportInfo {
			ErrorCode type;
			std::wstring description;
			unsigned code;
			void *data;
		};

	public:
		/**
		@brief Initialize error handling/display singleton
		@param data Data needed for initialization (this is platform specific)
		*/

		static void Initialize(void *fs_client);

		/**
		@brief Deinitialize error handling/display singleton
		*/	
		static void Deinitialize(void);

		/**
		@brief Get instance pointer to error handling singleton
		@note Instance must be initialized explicitly before calling this
		@return Returns a handle to the instance of this class.
		*/
		static Error *GetInstance(void);


		/**
		@brief Load the errviewer RPL. Must be done after audio and video are initialized.
		*/
		void Load(void);

		void Unload(void);
		/**
		@brief Create Custom Error List - part of constructor, untested if called outside
		*/
		void CreateCustomErrorList(unsigned, unsigned);

		/**
		@brief Change how fatal errors are handled.
		*/
		void ChangeFatalErrorHandling(bool);

		/**
		@brief Sets the visibility of different error types.

		    Each mode shows a subset of the errors the one above it shows.
		    CORE - Errors that should be visible to platform developers
		    APP - Errors that should be visible to application developers
		    USER - Errors that should be visible to users of applications
		    NONE - Display no errors graphically
		
		@param errormode A string that dictates which class of errors are visibly shown to the user
		*/
		void SetMode(const std::string & errormode);

		/**
		@brief Recieve messages pertaining to error UI state changes
		@param callback This function will get called whenever there is a state change to the error UI
		*/
		void AddGlobalCallback(StateChangeCallback callback);
		
		/**
		@brief Invoke the error UI with the following message...
		@note Leaving all text parameters blank will result in a default UI message with a single button
			The UI will only display two buttons if text is given for the right_button_text param
		@param main_text Text shown in main window
		@param left_button_text Text shown on the left most button
		@param right_button_text Text shown on the right most button
		*/
		#ifdef PF_UNIXBLD
		// Note: On unix, UChar and wchar_t are _NOT_ the same
		void IAmError(
			Visibility visibility = CORE,
			const UChar *main_text = 0,
			const UChar *left_button_text = 0,
			const UChar *right_button_text = 0,
			const UChar *title_text = 0,
			LibPlatform::Error::ScreenType screenType = DISPLAY_ALL,
			bool shouldPause = true
		) {
			// Fixme
		}
		#endif
		void IAmError(
			Visibility visibility = CORE,
			const wchar_t *main_text = 0, 
			const wchar_t *left_button_text = 0, 
			const wchar_t *right_button_text = 0,
			const wchar_t *title_text = 0,
			LibPlatform::Error::ScreenType screenType = DISPLAY_ALL,
			bool shouldPause = true
		);

		/**
		@brief Invoke the error UI with 'baked' message
		@note
		@param error_type What type of error do you want to display?s
		*/
		void IAmError(ErrorCode error_type, bool shouldPause = true, LibPlatform::Error::ScreenType screenType = DISPLAY_ALL);

		/**
		@brief Tells the HandleInput loop that there has been a fatal error 
		*/
		void setFatalError(void (*doneCallback)(void)); 
		
		/**
		@brief Sets the active controller by index
		*/
		void setActiveController(int index);

		/**
		@brief Sets the relaunch args for when the application is going to restart
		@param count The number of command line arguments
		@param args The command line arguments for the application
		*/
		void setRelaunchArgs(int count, char **args);

		enum PostErrorAction
		{
			eAct_Nothing,
			eAct_ReturnToMenu,
			eAct_ReturnToNetworkSettings,
			eAct_ReturnToDataManagement,
			eAct_RelaunchApplication
		};

		struct PostErrorUserCallback
		{
			PostErrorUserCallback() : mpUserCallback(0), mpUserValue(0) { }
			PostErrorAction	(*mpUserCallback)(int dialogButton, void *pUserValue);
			void			*mpUserValue;
		};

		/**
		@brief Invoke the error UI with 'baked' message
		*/
		void EnqueueError(unsigned int errviewerCode, PostErrorAction postAction, const PostErrorUserCallback *pOptCallback);

		/**
		@brief Invoke the error UI with user defined message
		*/
		void EnqueueError(Visibility visibility, const wchar_t *main_text, const wchar_t *left_button_text, const wchar_t *right_button_text, const wchar_t *title_text, ScreenType screenType,
						  PostErrorAction postAction, const PostErrorUserCallback *pOptCallback);

		/**
		@brief Closes the Dialog window without pressing a button, also flushes the error queue
		*/
		void CloseDialog();

		/**
		@brief Let the gui know the state of the remote
		@param kpad_status_pointer State of input device
		@param channel device channel
		*/
		void SetKPADInfoPointer(void *kpad_status_pointer, int channel);
		
		/**
		@brief Let the gui know the state of the remote
		@param vpad_status_pointer State of input device
		*/
		void SetVPADInfoPointer(void *vpad_status_pointer);
		
		/**
		@brief Signals the error gui to update its state based on the provided input from SetKPADInfoPointer and SetVPADInfoPointer
		@return Returns 0 if the error was a non-fatal error.  -1 if the error as a non-fatal error. 
		*/
		int HandleInput(void);
		
		/**
		@brief Render the gui to the TV
		*/
		void RenderTV(void *pColorBuffer);

		/**
		@brief Render the gui to the GamePad
		*/
		void RenderGamePad(void *pColorBuffer);

		/**
		@brief DisappearHomeNixSign
		*/
		void HideHBMDeniedIcon(void);
		
		/**
		@brief Is the gui currently showing?
		@return Returns true if the state of the error gui is to be visible
		*/
		bool IsShowing(void);

		/**
		@brief Report an error within the system. 
			This error may or may not every graphically be seen by the user.
			This system will check all reported errors and attempt to handle them itself (if it can).
			Whether or not a message is shown graphically or not is on a case by case basis
		@param type What type of error we're dealing with
		@param description Brief explanation of error
		@param data Any data needed for handling the error
		*/
		#ifdef PF_UNIXBLD
		// Note: On unix, UChar and wchar_t are _NOT_ the same
		void Report(ErrorCode type, const UChar *description, void *data) { }
		#endif
		void Report(ErrorCode type, std::wstring description, void *data);

		/**
		@brief For checking the state of the systems errors
		@return Returns false if no errors have been reported
		*/
		bool HasReports(void);

		/**
		@brief Retrieve the info for a reported error
		@return Returns the next report from the queue
		*/
		ReportInfo NextReport(void);

		/**
		@brief Convert CURL error code to ErrorViewer compatible error code
		@return Returns error code for viewer or zero if no matching code could be found
		*/
		static ErrorCode GetErrorCodeFromCurlError(int curlError);

		/**
		@brief Returns the offset at which HTTP status codes begin
		@return Returns the offset at which HTTP status codes begin
		*/
		static ErrorCode GetHTTPStatusCodesOffset(){return ERROR_CODE_WEBKIT_STATUS_CODES_BEGIN;}

	private:
		// Singleton :/
		Error(void *fs_client);
		~Error(void);
		int m_activeController;
		bool m_fatalError;
		void (*m_fatalErrorCB)(void);
		bool m_fatalMessageTriggered;
		
		static PostErrorAction FatalErrorDoneCallback(int dialogButton, void *pUserValue);
		std::map<ErrorCode, CustomErrorData> customWebKitErrorMap;
	};
}

extern "C" {
void SetErrorOutputFunc(void (*funcP)(void));
};
#endif//ERROR_H
